@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: antic.cpp,v 1.110 2013-01-02 20:56:14 thor Exp $
+ ** $Id: antic.cpp,v 1.127 2015/12/11 16:27:35 thor Exp $
  **
  ** In this module: Antic graphics emulation
  **
@@ -171,13 +171,14 @@ Antic::Antic(class Machine *mach)
     ModeLines[i] = NULL;
   //
   // Reset the current modeline generator
-  CurrentMode    = NULL;
+  ScanlineGenerator.NoMode();
   //
   GTIAStart           = 16; // Start position of GTIA.
   YPosIncSlot         = 111;
   YPos                = 0;
   //
   NTSC                = false;
+  isAuto              = true;
   TotalLines          = PALTotal;
   PreviousIR          = 0x0; // Blank lines
   //
@@ -321,7 +322,7 @@ void Antic::WarmStart(void)
   // Reset the internal Antic state machine
   //
   for(i = 0;i < 5;i++) {
-    BusNoise[i] = UBYTE(rand() >> 8);
+    PlayerData[i] = 0x00;
   }
   AnticPC        = 0x00;
   AnticPCShadow  = 0x00;
@@ -333,10 +334,11 @@ void Antic::WarmStart(void)
   CharCtrl       = 0x00;
   HScroll        = 0x00;
   VScroll        = 0x00;
-  DMACtrlShadow  = 0x00; 
   PreviousIR     = 0x00; // Blank lines 
+  //
   // Reset the current modeline generator
-  CurrentMode    = NULL;
+  ScanlineGenerator.NoMode();
+  //
   if (NTSC) {
     TotalLines = NTSCTotal;
   } else {
@@ -380,34 +382,36 @@ void Antic::ModeLine2::Generator(ULONG *ptr,int width,int scanline)
 				    PACK4(GPFF,GPF2,GPFF,GPF2),PACK4(GPFF,GPF2,GPFF,GPFF),
 				    PACK4(GPFF,GPFF,GPF2,GPF2),PACK4(GPFF,GPFF,GPF2,GPFF),
 				    PACK4(GPFF,GPFF,GPFF,GPF2),PACK4(GPFF,GPFF,GPFF,GPFF) };
-  ADR chbase;
   int nchars  = width >> 3;
-  const UBYTE *scan = ScanBuffer;
-  
-  scanline &= 7;
+  const UBYTE *scan = ScanBuffer; 
+  static const UBYTE OffsetNormal[16] = {0,1,2,3,4,5,6,7,8,8,2,3,4,5,6,7};
+  static const UBYTE OffsetLow[16]    = {0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7};
+  static const UBYTE *const LineGenerator[8] = {OffsetNormal,OffsetNormal,OffsetNormal,OffsetLow,
+						OffsetNormal,OffsetNormal,OffsetNormal,OffsetLow};
+ 
+  scanline &= 0x0f;
 
-  if (!CharGen->UpsideDown) {
-    chbase = CharGen->CharBase + scanline;
-  } else {
-    chbase = CharGen->CharBase + 7 - scanline;
-  }
-  
   do {  
-    ADR   chaddr;
+    ADR   chaddr,chline;
     UBYTE chdata,screendata;
     // get screen data from antic buffer
-    screendata = *scan++;  
-    // get character base address
-    chaddr     = chbase + ((screendata & 0x7f) << 3);  
-    // get character data
-    chdata     = UBYTE(CharGen->Ram->ReadByte(chaddr));
+    screendata = *scan++; 
+    chline     = LineGenerator[screendata >> 5][scanline];
+    // Check whether we have a completely blank line. If so, supply zero
+    if (chline >= 8) {
+      chdata = 0;
+    } else {
+      chline ^= CharGen->UpsideDown;
+      chaddr  = CharGen->CharBase + chline + ((screendata & 0x7f) << 3);
+      chdata  = UBYTE(CharGen->Ram->ReadByte(chaddr));
+    }
+    
+    if (screendata & CharGen->BlankMask)
+      chdata  = 0;
     
     if (screendata & CharGen->InvertMask)
       chdata ^= 0xff;
 
-    if (screendata & CharGen->BlankMask)
-      chdata  = 0;
-    
     *ptr++ = Lut[chdata >> 4].l;
     *ptr++ = Lut[chdata & 0x0f].l;
     
@@ -429,41 +433,33 @@ void Antic::ModeLine3::Generator(ULONG *ptr,int width,int scanline)
 				    PACK4(GPFF,GPFF,GPFF,GPF2),PACK4(GPFF,GPFF,GPFF,GPFF) };
   int nchars  = width >> 3;
   const UBYTE *scan = ScanBuffer;
-  static const UBYTE OffsetNormal[10] = {0,1,2,3,4,5,6,7,8,8};
-  static const UBYTE OffsetLow[10]    = {8,8,2,3,4,5,6,7,0,1};
+  static const UBYTE OffsetNormal[16] = {0,1,2,3,4,5,6,7,8,8,2,3,4,5,6,7};
+  static const UBYTE OffsetLow[16]    = {8,8,2,3,4,5,6,7,0,1,2,3,4,5,6,7};
+  static const UBYTE *const LineGenerator[8] = {OffsetNormal,OffsetNormal,OffsetNormal,OffsetLow,
+						OffsetNormal,OffsetNormal,OffsetNormal,OffsetLow};
 
-  if (scanline > 10)
-    scanline &= 7;
+  scanline &= 0x0f;
  
   do {  
     ADR   chaddr,chline;
     UBYTE chdata,screendata;
     // get screen data from antic buffer
-    screendata = *scan++;    
-    if ((screendata & 0x60) == 0x60) {
-      // This is a lowercase character. This means that the first two
-      // scan lines are blank and get mirrored as the last two scan lines
-      chline = OffsetLow[scanline];
-    } else {
-      chline = OffsetNormal[scanline];
-    }
+    screendata = *scan++;
+    chline     = LineGenerator[screendata >> 5][scanline];    
     // Check whether we have a completely blank line. If so, supply zero
     if (chline >= 8) {
       chdata = 0;
     } else {
-      if (!CharGen->UpsideDown) {
-	chaddr = CharGen->CharBase + chline       + ((screendata & 0x7f) << 3);
-      } else {
-	chaddr = CharGen->CharBase + 7 - chline   + ((screendata & 0x7f) << 3);
-      }
-      chdata   = UBYTE(CharGen->Ram->ReadByte(chaddr));
+      chline ^= CharGen->UpsideDown;
+      chaddr  = CharGen->CharBase + chline + ((screendata & 0x7f) << 3);
+      chdata  = UBYTE(CharGen->Ram->ReadByte(chaddr));
     }
     
-     if (screendata & CharGen->InvertMask)
-      chdata ^= 0xff;
-
     if (screendata & CharGen->BlankMask)
       chdata  = 0;
+
+    if (screendata & CharGen->InvertMask)
+      chdata ^= 0xff;
     
     *ptr++ = Lut[chdata >> 4].l;
     *ptr++ = Lut[chdata & 0x0f].l;
@@ -483,17 +479,14 @@ void Antic::ModeLine4::Generator(ULONG *ptr,int width,int scanline)
 				      PACK2(GPF0,GPFB), PACK2(GPF0,GPF0), PACK2(GPF0,GPF1), PACK2(GPF0,GPF3),
 				      PACK2(GPF1,GPFB), PACK2(GPF1,GPF0), PACK2(GPF1,GPF1), PACK2(GPF1,GPF3),
 				      PACK2(GPF3,GPFB), PACK2(GPF3,GPF0), PACK2(GPF3,GPF1), PACK2(GPF3,GPF3) };
+  static const BytePack *const Luts[2] = {LutLo,LutHi};
   ADR chbase;
   int nchars  = width >> 3;
   const UBYTE *scan = ScanBuffer;
 
-  scanline &= 7;
-  
-  if (!CharGen->UpsideDown) {
-    chbase = CharGen->CharBase + scanline;
-  } else {
-    chbase = CharGen->CharBase + 7 - scanline;
-  }
+  scanline &= 7; 
+  scanline ^= CharGen->UpsideDown;
+  chbase    = CharGen->CharBase + scanline;
   
   do {  
     ADR   chaddr;
@@ -506,9 +499,7 @@ void Antic::ModeLine4::Generator(ULONG *ptr,int width,int scanline)
     chaddr     = chbase + ((screendata & 0x7f) << 3);  
     // get character data
     chdata     = CharGen->Ram->ReadByte(chaddr);
-    
-    lut        = (screendata & 0x80)?(LutHi):(LutLo);
-    
+    lut        = Luts[screendata >> 7];
     *ptr++     = lut[chdata >> 4].l;
     *ptr++     = lut[chdata & 0x0f].l;
   } while(--nchars);
@@ -527,15 +518,14 @@ void Antic::ModeLine5::Generator(ULONG *ptr,int width,int scanline)
 				      PACK2(GPF0,GPFB), PACK2(GPF0,GPF0), PACK2(GPF0,GPF1), PACK2(GPF0,GPF3),
 				      PACK2(GPF1,GPFB), PACK2(GPF1,GPF0), PACK2(GPF1,GPF1), PACK2(GPF1,GPF3),
 				      PACK2(GPF3,GPFB), PACK2(GPF3,GPF0), PACK2(GPF3,GPF1), PACK2(GPF3,GPF3) };
+  static const BytePack *const Luts[2] = {LutLo,LutHi};
   ADR chbase;
   int nchars  = width >> 3;
   const UBYTE *scan = ScanBuffer;
   
-  if (!CharGen->UpsideDown) {
-    chbase = CharGen->CharBase + (scanline >> 1);
-  } else {
-    chbase = CharGen->CharBase + 7 - (scanline >> 1);
-  }
+  scanline >>= 1;
+  scanline  ^= CharGen->UpsideDown;
+  chbase     = CharGen->CharBase + scanline;
   
   do {  
     ADR   chaddr;
@@ -548,9 +538,7 @@ void Antic::ModeLine5::Generator(ULONG *ptr,int width,int scanline)
     chaddr     = chbase + ((screendata & 0x7f) << 3);  
     // get character data
     chdata     = CharGen->Ram->ReadByte(chaddr);
-    
-    lut        = (screendata & 0x80)?(LutHi):(LutLo);
-    
+    lut        = Luts[screendata >> 7];
     *ptr++     = lut[chdata >> 4].l;
     *ptr++     = lut[chdata & 0x0f].l;
   } while(--nchars);
@@ -571,12 +559,8 @@ void Antic::ModeLine6::Generator(ULONG *ptr,int width,int scanline)
   const UBYTE *scan = ScanBuffer;
 
   scanline &= 7;
-  
-  if (!CharGen->UpsideDown) {
-    chbase = CharGen->CharBase + scanline;
-  } else {
-    chbase = CharGen->CharBase + 7 - scanline;
-  }
+  scanline ^= CharGen->UpsideDown;
+  chbase    = CharGen->CharBase + scanline;
   
   do {  
     const BytePack *lut;
@@ -591,12 +575,11 @@ void Antic::ModeLine6::Generator(ULONG *ptr,int width,int scanline)
     chdata     = CharGen->Ram->ReadByte(chaddr);
     // get cell color
     lut        = Luts[screendata >> 6];
-
-    ptr[3] = lut[chdata & 0x03].l,chdata >>= 2;
-    ptr[2] = lut[chdata & 0x03].l,chdata >>= 2;
-    ptr[1] = lut[chdata & 0x03].l,chdata >>= 2;
-    ptr[0] = lut[chdata].l;
-    ptr   += 4;
+    ptr[3]     = lut[chdata & 0x03].l,chdata >>= 2;
+    ptr[2]     = lut[chdata & 0x03].l,chdata >>= 2;
+    ptr[1]     = lut[chdata & 0x03].l,chdata >>= 2;
+    ptr[0]     = lut[chdata].l;
+    ptr       += 4;
 
   } while(--nchars);
 }
@@ -615,11 +598,9 @@ void Antic::ModeLine7::Generator(ULONG *ptr,int width,int scanline)
   int nchars  = width >> 4;
   const UBYTE *scan = ScanBuffer;
   
-  if (!CharGen->UpsideDown) {
-    chbase = CharGen->CharBase +     (scanline >> 1);
-  } else {
-    chbase = CharGen->CharBase + 7 - (scanline >> 1);
-  }
+  scanline >>= 1;
+  scanline  ^= CharGen->UpsideDown;
+  chbase     = CharGen->CharBase + scanline;
   
   do {  
     const BytePack *lut;
@@ -784,15 +765,62 @@ void Antic::ModeLineF::Generator(ULONG *ptr,int width,int)
 }
 ///
 
+/// Antic::Scanline::ComputeLineParameters
+// Precompute the parameters for generating a single
+// scan line. First argument is the modeline,
+// then the active DMA settings for the scrolled screen,
+// followed by the regular parameters defining the nominal
+// borders where we clip.
+// Last are the output buffer itself, followed by the scroll
+// offset and the current line in the modeline.
+void Antic::Scanline::ComputeLineParameters(struct ModeLine *mode,
+					    struct DMAGenerator *dma,
+					    struct DMAGenerator *borders,
+					    UBYTE *buffer,int xscroll,int displayline)
+{ 
+  CurrentMode = mode;
+  Width       = dma->Playfield.NumCycles << 2;
+  XMin        = borders->FillInOffset + (Antic::FillInOffset - xscroll);
+  if (XMin < Antic::FillInOffset)
+    XMin = Antic::FillInOffset;
+  XMax        = XMin + (borders->Playfield.NumCycles << 2);
+  FillIn      = buffer + Antic::FillInOffset + dma->FillInOffset;
+  LineBuffer  = buffer;
+  DisplayLine = displayline;
+}
+///
+
+/// Antic::Scanline::GenerateScanline
+// Generate a single scan line.
+void Antic::Scanline::GenerateScanline(void) const
+{
+  if (LineBuffer) {
+    if (Width) {
+      CurrentMode->Generator((ULONG *)(FillIn),Width,DisplayLine);
+      //
+      // shift is the displacement from the position where the data did go
+      // to the position where it should have gone. GTIA corrects this, then.
+      //
+      // Antic generated too much display, we need to erase the display data
+      // that got shifted out of the frame. This requires handling of the
+      // special case where the display is wide already. Then, more zeros
+      // have to enter.
+      memset(LineBuffer     ,GTIA::Background,XMin);
+      memset(LineBuffer+XMax,GTIA::Background,Antic::DisplayModulo-XMax);
+    } else {
+      memset(LineBuffer     ,GTIA::Background,Antic::DisplayModulo);
+    }
+  }
+}
+///
+
 /// Antic::HBI
 // Generate a horizontal blank.
 void Antic::HBI(void)
 {
-  // Update DMACtrl for player DMA
 }
 ///
 
-
 /// Antic::FetchPlayfield
 // Load data from the playfield into the scanline buffer
 void Antic::FetchPlayfield(struct ModeLine *mode,struct DMAGenerator *dma)
@@ -820,73 +848,46 @@ void Antic::FetchPlayfield(struct ModeLine *mode,struct DMAGenerator *dma)
 }
 ///
 
-/// Antic::Scanline
-// Generate one scanline of the Antic display
-// If nmi is true, then a DLI has to generated.
-// modeline is the mode line generator for this scanline.
-// fillin is where the playfield data shall go. It may be offset to screen.
-// width is the number of pixels the modeline shall fill in, shift the
-// shift offset. 
-// displayline is the number of the scanline within the modeline this method
-// generates.
-// Emulation here starts at value zero of the "horizontal register" as defined
-// by the technical manual of Antic, sheet 5. The horizontal register itself
-// counts half color clocks, similar to to our pixel based emulation.
-// Measurements show that the DLI reaches the CPU about 8 cycles after the
-// STA WSYNC position at hpos = 208. That fits well to the hypothesis that
-// a NMI is generated at hpos = 0 since we have 228 half color clocks
-// and well to the information in DeReAtari where it is stated that the
-// STA WSYNC must reach antic before clock cycle 103.
-//
-// This is all tricky. Some applications worth testing with:
-// Rescue, Nautilus, JetBoot, PitfallII, Zaxxon, Frogger, Decathlon,
-// Koronis Rift, Moon Patrol, Fort Apocalypse
-void Antic::Scanline(struct ModeLine *mode,
-		     UBYTE *fillin,int width,int xscroll,int displayline)
+/// Antic::FetchPlayerMissiles
+// Fetch the player-missile graphics
+void Antic::FetchPlayerMissiles(void)
 {
-  bool wasfiddled = false;
-  // Keep scrolling data because it may change under our feed since the
-  // CPU may change it while we're working
-  // shift is the displacement from the position where the data did go
-  // to the position where it should have gone. GTIA corrects this, then.
-  int xmin,xmax,shift;
+  int i;
+
   //
-  shift = FillInOffset - xscroll;
-  xmin  = ActiveDMATiming->Regular.FillInOffset + shift;
-  xmax  = xmin + (ActiveDMATiming->Regular.Playfield.NumCycles << 2);
-  //
-  // Now generate the display if there is someting to set...
-  // Keep the data to be able to re-run the generator in case someone sets up a horizontal kernel
-  // modifying antic data.
-  //
-  // Keep the fiddling flag from the last line to emulate the lost-sync bug if the last
-  // line in the display is a fiddled mode.
-  if (CurrentMode)
-    wasfiddled = CurrentMode->Fiddling;
-  CurrentMode = mode;
-  //
-  if (width) {
-      mode->Generator((ULONG *)(fillin),width,displayline);
-      //
-      // shift is the displacement from the position where the data did go
-      // to the position where it should have gone. GTIA corrects this, then.
-      //
-      // Antic generated too much display, we need to erase the display data
-      // that got shifted out of the frame. This requires handling of the
-      // special case where the display is wide already. Then, more zeros
-      // have to enter.
-      if (xmin < FillInOffset) xmin = FillInOffset;
-      memset(LineBuffer     ,GTIA::Background,xmin);
-      memset(LineBuffer+xmax,GTIA::Background,DisplayModulo-xmax);
+  // Read player data.
+  if (DMACtrl & 0x08) {
+    for(i = 0;i < 4;i++) {
+      PlayerData[i] = UBYTE(Ram->ReadByte(PMActive->PlayerBase[i] 
+					  + (YPos >> PMActive->YPosShift)));
+    }
+  } else if (DMACtrl & 0x20) {
+    // Simulate bus noise
+    for(i = 0;i < 4;i++) {
+      PlayerData[i] = UBYTE(rand() >> 8);
+    }
   } else {
-      memset(LineBuffer     ,GTIA::Background,DisplayModulo);
+    // BUNDES.BAS seems to forget to clear GRACTL or picks, by pure chance,
+    // 0 as bus noise??
+    PlayerData[4] = 0;
+    for(i = 0;i < 4;i++) {
+      PlayerData[i] = 0;
+    }
+  }
+  //
+  // Read missile data.
+  // This is enabled whenever player *or* missile data is on.
+  if (DMACtrl & 0x0c) {
+    PlayerData[4] = UBYTE(Ram->ReadByte(PMActive->MissileBase
+					+ (YPos >> PMActive->YPosShift)));
+  } else if (DMACtrl & 0x20) {
+    // Simulate bus noise
+    PlayerData[4] = UBYTE(rand() >> 8);
+  } else {
+    // BUNDES.BAS seems to forget to clear GRACTL or picks, by pure chance,
+    // 0 as bus noise??
+    PlayerData[4] = 0;
   }
-  // 
-  // Run now the display generator.
-  // Antic has a bug here that it even drives GTIA (and does not generate a sync pulse)
-  // if the last line is a hires line.
-  if (YPos < DisplayHeight || wasfiddled)
-    Gtia->TriggerGTIAScanline(LineBuffer+shift,0,DisplayModulo - FillInOffset,mode->Fiddling);
 }
 ///
 
@@ -908,52 +909,53 @@ void Antic::Modeline(int ir,int vertscroll,int nlines,struct ModeLine *gen)
   int shift    = 0;
   int scanline = 0;
   int dmadelta;             // DMA cycle shift allocation due to HScroll
-  bool isfirst = true;
-  bool islast  = false;
+  bool isfirst    = true;
+  bool islast     = false;
+  bool wasfiddled = false;
   
   nlines--; // makes a couple of things simpler.
   //
   // Note that the test for "end of mode" is done here again
   // and that this decision depends on a potentially updated version
-  // of VScroll.
-  while(scanline <= ((vertscroll == 1)?(VScroll):(nlines)) && YPos <= DisplayHeight) {
-    UBYTE *fillin;
-    int firstcycle;
-    //
+  // of VScroll. Note that VScroll must be sampled at the start of the line
+  // for computing the first line, but not the last line.
+  scanline = (vertscroll == 2)?(VScroll):0;
+  while(islast == false && YPos <= DisplayHeight) {
     // First reserve P/M DMA slots. They appear first because
     // The CPU needs to see them before the DLI triggers in.
-    switch (DMACtrl & 0x0c) {      
-    case 0x08: // Player DMA
-    case 0x0c: // Missile DMA is for free if player DMA is turned on.
-      Cpu->StealCycles(PlayerFetchSlot);
-      // runs intentionally into the missile fetch
-    case 0x04: // Missile DMA
-      Cpu->StealCycles(MissileFetchSlot);
-      break;
+    if (YPos < DisplayHeight) {
+      if (DMACtrl & 0x08) {
+	// Player DMA is enabled. Steal both the player DMA cycle
+	// as well as the missle DMA cycle, no matter whether
+	// missile DMA is on.
+	Cpu->StealCycles(PlayerFetchSlot);
+	Cpu->StealCycles(MissileFetchSlot);
+      } else if (DMACtrl & 0x04) {
+	// Missle DMA only.
+	Cpu->StealCycles(MissileFetchSlot);
+      }
+      // Fetch now the P/M data for GTIA (or not).
+      FetchPlayerMissiles();
     }
-    // Keeps the information on which P/M data will be made available
-    // to GTIA.
-    DMACtrlShadow = DMACtrl;    
-    // 
+    //
     // Advance the CPU for a couple of cycles before the DLI is triggered (Jetboot Jack)
     // and before the DMA settings become active.
-    Cpu->Go(6); 
+    Cpu->Go(6);
     //
-    if (isfirst && vertscroll == 2) {
-      // This is the first vertical scrolling line, last was regular
-      // Hence, delay the first line.
-      scanline = VScroll;
-      if (scanline > nlines)
-	scanline -= 16;
-    } 
+    // While the vertical line start is sampled at the start of the 
+    // scanline, the end of the modeline is sampled later.
+    // Moving this earlier breaks the acid test, moving the vscroll
+    // start breaks numen.
+    int last = (vertscroll == 1)?(VScroll):(nlines);
     //
     // Check whether this is the last line of the mode and we should
-    // possibly generate a DLI.
-    if (scanline == ((vertscroll == 1)?(VScroll):(nlines)))
+    // possibly generate a DLI. This is more elegant than the
+    // previous counter adjustment.
+    if (((scanline ^ last) & 0x0f) == 0)
       islast = true;
     // 
     // Advance to the DLI-generation step of ANTIC.
-    Cpu->Step(); 
+    Cpu->Step();
     //
     // If this is the last scanline and the "generate DLI" bit of the
     // instruction is set, and DLIs are allowed, signal that we need
@@ -976,14 +978,12 @@ void Antic::Modeline(int ir,int vertscroll,int nlines,struct ModeLine *gen)
       NMIStat = (YPos == VBIStart)?(0x40):(0x80);
       //
       // But the CPU can enable the NMI.
-      if ((NMIEnable & NMIStat) & 0xc0) {
+      if (((NMIEnable & NMIStat) & 0xc0) && (nmi == false)) {
 	// Generate now if NMI is not yet active, but again give the CPU some time.
-	if (!nmi) {
-	  Cpu->Step();
-	  nmi = true;
-	}
+	Cpu->Step();
+	nmi = true;
+	// Note that the extra step taken here is compensated below.
       }
-      //
       Cpu->Go(2);
       //
       if (nmi) {
@@ -993,7 +993,7 @@ void Antic::Modeline(int ir,int vertscroll,int nlines,struct ModeLine *gen)
     }
     //
     // Compute the number of clocks to the start of the display - DMA.
-    firstcycle       = GTIAStart; // At worst here because then GTIA starts.
+    int firstcycle   = GTIAStart; // At worst here because then GTIA starts.
     if (ir & 0x10) {
       dmadelta       = HScroll >> 1;
       dma            = &(ActiveDMATiming->Scrolled);
@@ -1027,19 +1027,14 @@ void Antic::Modeline(int ir,int vertscroll,int nlines,struct ModeLine *gen)
     }
     //
     // Now compute the DMA timings, finally.
-    // Displace data by the fill-in offset for easier horizontal scrolling.
-    // We can scroll at most 16 color clocks, hence a fill-in of 32 half
-    // color clocks.
     if (ir & 0x10) {
       // DMA cycle shifting is now a bit tricky: One DMA cylce is two color clocks.
       dmadelta       = HScroll >> 1; // shift by half the cycles to the right
       // Computations change if horizontal scrolling is enabled
-      fillin         = LineBuffer   + FillInOffset + ActiveDMATiming->Scrolled.FillInOffset;
       shift          = HScroll << 1;
       dma            = &(ActiveDMATiming->Scrolled);
     } else {
       dmadelta       = 0;
-      fillin         = LineBuffer   + FillInOffset + ActiveDMATiming->Regular.FillInOffset;
       shift          = 0;
       dma            = &(ActiveDMATiming->Regular);
     }
@@ -1063,7 +1058,7 @@ void Antic::Modeline(int ir,int vertscroll,int nlines,struct ModeLine *gen)
       // cycles required to fetch the instruction
       if (isfirst) {
 	if (gen->FontCycles) {     
-	  // Add cycles for fetching the character glyphcs.
+	  // Add cycles for fetching the character glyphs.
 	  gfx.FirstCycle = dma->Glyph.FirstCycle + dmadelta;
 	  gfx.NumCycles  = dma->Glyph.NumCycles;
 	  gfx.CycleMask  = gen->DMACycles;
@@ -1075,6 +1070,7 @@ void Antic::Modeline(int ir,int vertscroll,int nlines,struct ModeLine *gen)
 	  gfx.CycleMask  = gen->DMACycles;
 	}
 	//
+	// Now get the data for which the cycles were requested.
 	if (gen->DMAShift && dma->Playfield.NumCycles) {
 	  gfx.LastCycle  = 106;
 	  Cpu->StealCycles(gfx);
@@ -1086,21 +1082,49 @@ void Antic::Modeline(int ir,int vertscroll,int nlines,struct ModeLine *gen)
     // Get the base DMA cycles for character based modes. This
     // part of the DMA fetches characters into the mode line buffer.
     // Include memory refresh. This is additional for all scan lines, not
-    // only for the first scan line of a mode line. Note that we are here
-    // running in a loop.
+    // only for the first scan line of a mode line. 
     Cpu->StealMemCycles(MemRefreshSlot);
     //
     // Advance the CPU to the first GTIA cycle.
-    if (GTIAStart - firstcycle > 0)
+    if (GTIAStart > firstcycle)
       Cpu->Go(GTIAStart - firstcycle);
     //
-    // Now run the scanline of this mode line
-    Scanline(gen,fillin,dma->Playfield.NumCycles << 2,shift,scanline);
+    // Keep the fiddling flag from the last line to emulate the lost-sync bug if the last
+    // line in the display is a fiddled mode.
+    wasfiddled = ScanlineGenerator.isFiddled();
+    //
+    // Setup the parameters for the scanline thus
+    // we can regenerate if CHARCTL or some other mode
+    // parameter changes on the screen.
+    ScanlineGenerator.ComputeLineParameters(gen,dma,&(ActiveDMATiming->Regular),
+					    LineBuffer,shift,scanline);
+    //
+    // And run it.
+    ScanlineGenerator.GenerateScanline();
+    // 
+    // Run now the display generator.
+    // Antic has a bug here that it even drives GTIA (and does not generate a sync pulse)
+    // if the last line is a hires line.
+    if (YPos < DisplayHeight || wasfiddled) {
+      // Create the GTIA output, player/missile, priority logic
+      // plus GTIA mode processing. This also runs the CPU for the
+      // remaining cycles in the playfield.
+      Gtia->TriggerGTIAScanline(LineBuffer + FillInOffset - shift,PlayerData,
+				DisplayModulo - FillInOffset,gen->Fiddling);
+    } else {
+      // Otherwise, dry-run of the CPU. Note that if there are more cycles
+      // requested here than fit into the horizonal line, the remaining
+      // cycles are no-ops. The CPU is not advanced for these extra cycles
+      // and the machine state is not advanced.
+      Cpu->Go((DisplayModulo - FillInOffset) >> 2);
+    }
     //
     // We are now on the next line.
     YPos++;
     scanline++;
     isfirst = false;
+    //
+    // Signal start of the next line.
     machine->HBI();
   }
 }
@@ -1121,13 +1145,21 @@ void Antic::RunDisplayList(void)
   if ((currentir & 0x0f) >= 2 && (currentir & 0x020)) { // a display mode?
     vertscroll = 2;
   }
+  if (currentir & 0x0f) {
+    // This is not a blank-line mode. Remove the "vertical blank" flag
+    // and the LMS flag, both are not repeated.
+    currentir &= ~0x40;
+    // If this is a jump instruction, convert to blank lines
+    if ((currentir & 0x0f) == 1)
+      currentir = 0x00;
+  }
   // 
   // VCount must equal zero for some games but the first line
   // starts when vcount == 4. Hence, we process here vcount = 0..3
-  //
   YPos         = 0;
   AnticPCShadow= AnticPC;
   AnticPCCur   = AnticPC;
+  //
   // Also reset the display frontend now: Signal that we restart from top.
   machine->Display()->ResetVertical();
   //
@@ -1142,7 +1174,6 @@ void Antic::RunDisplayList(void)
     machine->HBI();
   } while(++YPos < DisplayStart);
 
-  //NMIStat = 0x00; // clear NMI status: Not really done at all....
   do {
     // Update vertical scrolling flags from last line
     vertscroll >>= 1;
@@ -1150,30 +1181,30 @@ void Antic::RunDisplayList(void)
     if (jvb) { 
       // We are not displaying anything: If the CPU does character generator modifications
       // now, no immediate consequences will be drawn.
-      CurrentMode = NULL;
+      ScanlineGenerator.NoMode();
       // Yes, just generate a blank display. Requires 2 DMA cycles for memory,
-      //  blank modeline, full display width, scanline #0.
-      // The wait VSYNC instruction does generate DLIs.
+      // blank modeline, full display width, scanline #0.
+      // The wait VSYNC instruction does generate DLIs if bit 7 is set.
+      // Use the blank modeline generator for the purpose of generating
+      // the right output.
       Modeline(currentir,0,1,ModeLines[0]);
     } else {
       struct ModeLine *gen; // mode line generator
-      int i,ir,nlines;      // Antic instruction register, # scan lines in the mode
-      //
-      // Update the Antic bus noise now.
-      for(i = 0;i < 5;i++) {
-	BusNoise[i] = UBYTE(rand() >> 8);
-      }
+      int ir,nlines;        // Antic instruction register, # scan lines in the mode
       //
       // If display DMA is disabled, re-run the same instruction again.
       if (DMACtrl & 0x20) {
 	// Here: Antic is active, we must generate a display.
 	AnticPCCur = AnticPC;
-	currentir = ram->ReadByte(AnticPC);
+	currentir  = ram->ReadByte(AnticPC);
+	PreviousIR = currentir;
 	INCPC; // Increment the PC.
 	// Reserve the slots for DMA access for the DLI counter.
 	Cpu->StealCycles(DListFetchSlot);
-      } // DLI remains on, even if we re-fetch the IR.
+      } 
+      // DLI remains on, even if the same instruction is repeated.
       ir = currentir;
+      //
       // Now check for details about the instructions. We first
       // check for blank and jump instructions.
       if ((ir & 0x0f) == 0x00) {
@@ -1226,22 +1257,22 @@ void Antic::RunDisplayList(void)
       // Now go for the line generator.
       if (nlines > 0) {
 	// Everything else is displayed the regular way
-	//
 	Modeline(ir,vertscroll,nlines,gen);
       }
-    }    
+    } 
     // Note that the scanline function also increments the
     // vertical position and runs the pokey activity
   } while(YPos <= DisplayHeight);
   //
-  // Store IR for next go-around.
-  PreviousIR  = currentir;
-  //
   // Display list ended now, or we reached the end of the frame.
   // Generate now some blank lines for the vertical blank
   // We are not displaying anything: If the CPU does character generator modifications
   // now, no immediate consequences will be drawn.
-  CurrentMode = NULL;
+  ScanlineGenerator.NoMode();
+  //
+  // Note: Releasing NMI here breaks TWERPS because it uses a tight
+  // STA WSYNC loop that will extend beyond the current HPos. Leave it
+  // asserted as long as the vertical blank is active.
   //
   // Now run the vertical blank. The exact position depends on
   // whether we are a PAL or a NTSC machine
@@ -1299,13 +1330,19 @@ UBYTE Antic::NMIRead(void)
 // Define the character generator base address 
 void Antic::ChBaseWrite(UBYTE val)
 {
-  // First update the shadow, then modify the
-  // character generators
+  int lastcycle;
+
   ChBase          = ADR(val) << 8;
   // Note that we no longer modify the character generators here. This
   // happens at the end of the line.
   Char20.CharBase = ChBase & 0xfe00;
   Char40.CharBase = ChBase & 0xfc00; // Keep care of alignment restrictions
+  // 
+  // There should be a two-clock delay here...
+  lastcycle = ActiveDMATiming->Regular.Playfield.FirstCycle + 
+    ActiveDMATiming->Regular.Playfield.FirstCycle + 4;
+  if (Cpu->CurrentXPos() >= GTIAStart && Cpu->CurrentXPos() + 2 < lastcycle)
+    ScanlineGenerator.GenerateScanline();
 }
 ///
 
@@ -1313,12 +1350,20 @@ void Antic::ChBaseWrite(UBYTE val)
 // Write into the control register of the character generator
 void Antic::ChCtrlWrite(UBYTE val)
 { 
-  CharCtrl = val;
+  int lastcycle;
 
+  CharCtrl = val;
+  //
   // Set individual properties in the character generator
-  Char20.UpsideDown = Char40.UpsideDown = (val & 0x04)?(true):(false);
+  Char20.UpsideDown = Char40.UpsideDown = (val & 0x04)?(0x07):(0x00);
   Char20.InvertMask = Char40.InvertMask = (val & 0x02)?(0x80):(0x00);
   Char20.BlankMask  = Char40.BlankMask  = (val & 0x01)?(0x80):(0x00);  
+  //
+  // There should be a two-clock delay here...
+  lastcycle = ActiveDMATiming->Regular.Playfield.FirstCycle + 
+    ActiveDMATiming->Regular.Playfield.FirstCycle + 4;
+  if (Cpu->CurrentXPos() >= GTIAStart && Cpu->CurrentXPos() + 2 < lastcycle)
+    ScanlineGenerator.GenerateScanline();
 }
 ///
 
@@ -1327,7 +1372,16 @@ void Antic::ChCtrlWrite(UBYTE val)
 void Antic::DListLoWrite(UBYTE val)
 {
   AnticPC  = (AnticPC & 0xff00) | ADR(val);
+  //
+  // The shadow copy is just for the debugger.
   AnticPCShadow = AnticPC;
+  // AxisAssesin seems to require this, otherwise the DLI
+  // in the title screen rolls around forever.
+  //
+  // Actually, this is just a matter of bad luck
+  // as it depends on when precisely the intro notices that
+  // the START button is pressed.
+  //PreviousIR    = 0x00;
 }
 ///
 
@@ -1335,8 +1389,16 @@ void Antic::DListLoWrite(UBYTE val)
 // Write int the high-address of the display list counter
 void Antic::DListHiWrite(UBYTE val)
 {
-  AnticPC  = (AnticPC & 0x00ff) | (ADR(val) << 8);
+  AnticPC       = (AnticPC & 0x00ff) | (ADR(val) << 8);
   AnticPCShadow = AnticPC;
+  //
+  // AxisAssesin seems to require this, otherwise the DLI
+  // in the title screen rolls around forever.
+  //
+  // Actually, this is just a matter of bad luck
+  // as it depends on when precisely the intro notices that
+  // the START button is pressed.
+  // PreviousIR    = 0x00;
 }
 ///
 
@@ -1350,7 +1412,12 @@ void Antic::DMACtrlWrite(UBYTE val)
   // Compute the width of the playfield display from bits 0..1
   switch (val & 0x03) {
   case 0x00: // no playfield DMA
-    ActiveDMATiming       = &DMA_None;
+    ActiveDMATiming         = &DMA_None;
+    // Blank the display.
+    if (Cpu->CurrentXPos() >= GTIAStart) {
+      ScanlineGenerator.Width = 0;
+      ScanlineGenerator.GenerateScanline();
+    }
     break;
   case 0x01: // narrow playfield
     ActiveDMATiming       = &DMA_Narrow;
@@ -1396,6 +1463,8 @@ void Antic::VScrollWrite(UBYTE val)
 void Antic::NMIEnableWrite(UBYTE val)
 {
   NMIEnable = UBYTE(val | 0x1f); // Lower bits are forced to one
+  //
+  // Note that this does not trigger a pending NMI
 }
 ///
 
@@ -1436,6 +1505,10 @@ void Antic::PMBaseWrite(UBYTE val)
 void Antic::WSyncWrite(void)
 {
   // Stop the CPU until we reach the WSync position
+  //
+  // This pulls the RDY line. Note that this is different
+  // from HALT, which requests an immediate stop, and is
+  // used for DMA purposes.
   Cpu->WSyncStop();
 }
 ///
@@ -1508,16 +1581,32 @@ void Antic::ComplexWrite(ADR mem,UBYTE val)
 void Antic::ParseArgs(class ArgParser *args)
 {  
   static const struct ArgParser::SelectionVector videovector[] = 
-    { {"PAL" ,false},
-      {"NTSC",true },
+    { {"Auto",2    },
+      {"PAL" ,0    },
+      {"NTSC",1    },
       {NULL ,0}
     };
   LONG val;
 
-  val = NTSC;
+  val = NTSC?1:0;
+  if (isAuto)
+    val = 2;
   args->DefineTitle("ANTIC");
-  args->DefineSelection("VideoMode","sets ANTIC video mode",videovector,val);
-  NTSC = (val)?(true):(false);
+  args->DefineSelection("ANTICVideoMode","sets ANTIC video mode",videovector,val);
+  switch(val) {
+  case 0:
+    NTSC   = false;
+    isAuto = false;
+    break;
+  case 1:
+    NTSC   = true;
+    isAuto = false;
+    break;
+  case 2:
+    NTSC   = machine->isNTSC();
+    isAuto = true;
+    break;
+  }
   /*args->DefineLong("BeforeDLICycles",
     "set position of DLI in CPU cycles",2,11,BeforeDLICycles);*/
   /*args->DefineLong("YPosIncCycle",
@@ -1529,7 +1618,6 @@ void Antic::ParseArgs(class ArgParser *args)
   } else {
     TotalLines = PALTotal;
   }
-  //TotalLines     = (NTSC)?(NTSCTotal):(PALTotal);
 }
 ///
 
