@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: cpu.hpp,v 1.68 2013-01-03 12:01:09 thor Exp $
+ ** $Id: cpu.hpp,v 1.79 2021/08/16 10:31:01 thor Exp $
  **
  ** In this module: CPU 6502 emulator
  **********************************************************************************/
@@ -29,11 +29,6 @@ class Patch;
 
 /// Class CPU
 class CPU : public Chip, public Saveable, private HBIAction {
-#ifndef HAS_PRIVATE_ACCESS
-  // Compiler workaround: Substructures of a class are part of the class and have hence
-  // access to private data. This bug has been fixed in the GCC 3.xx series.
-public:
-#endif
   // The register set of the CPU follows here.
   UWORD GlobalPC;   // program counter
   UBYTE GlobalA;    // Accumulator
@@ -107,11 +102,6 @@ private:
     bool         ReleaseHalt;      // set by the main thread if WSync gets released.
   };
   //
-#ifndef HAS_PRIVATE_ACCESS
-  // Compiler workaround: Substructures of a class are part of the class and have hence
-  // access to private data. This bug has been fixed in the GCC 3.xx series.
-public:
-#endif
   // A pointer to the monitor 
   class Monitor       *monitor;
   //
@@ -163,10 +153,22 @@ public:
   // Pointer to the last available cycle.
   UBYTE *LastCycle;
   //
+  // In case this is non-NULL, it is a set of counters that is used for
+  // profiling code. Counters are increased as soon as the CPU
+  // reaches a given PC. There is one counter per memory location.
+  ULONG *ProfilingCounters;
+  //
+  // If this is non-NULL, it points to profiling counters that accumulate
+  // over subroutine calls.
+  ULONG *CumulativeCounters;
+  //
   // Current CPU cycle counter. This counts the number of cycles to the next
   // cycle reset and is used to control the sound output.
   ULONG  CycleCounter;
   //
+  // Profile counter, counts steps for the profiler.
+  ULONG  ProfileCounter;
+  //
   // CPU preferences
   LONG   WSyncPosition; // horizontal position of the WSync release slot. Defaults to 104.
   //
@@ -205,9 +207,21 @@ public:
   //
   // A wrapper for the above that supplies a virtual execution method.
   class MicroCode {
+    //
+    // The stop mask. Defines on which conditions this
+    // code can be halted. If 0x01, only on HALT, if 0x03, also on RDY.
+    UBYTE m_ucStopMask;
+    //
   public:
+    // Constructor: The argument is true (which should be the default)
+    // if the code should also stop on an RDY signal. The NMOS 6502
+    // does not do this on memory writes.
+    MicroCode(bool haltonrdy)
+      : m_ucStopMask((haltonrdy)?(0x03):(0x01))
+    { }
+    //
     virtual ~MicroCode(void)
-    {}
+    { }
     // The instruction step this execution unit performs
     // This is why we implement this, after all. It returns a value
     // required for the next execution unit. As registers are at most
@@ -218,6 +232,13 @@ public:
     // Insert another slot intermediately into the execution order. Might be called
     // dependent on the execution data.
     void Insert(class CPU *cpu);
+    //
+    // The stop mask. Can be 0x03 to stop on RDY (WSYNC) and HALT (DMA) or
+    // 0x01 to stop on HALT only. The latter goes for write instructions.
+    UBYTE StopMask(void) const
+    {
+      return m_ucStopMask;
+    }
   };
   //
   // The concatenation unit: This is a templated class that takes two basic instructions
@@ -228,8 +249,8 @@ public:
     Step2 Second;
     //
   public:
-    Cat2(class CPU *cpu)
-      : First(cpu), Second(cpu)
+    Cat2(class CPU *cpu, bool haltonrdy = true)
+      : MicroCode(haltonrdy), First(cpu), Second(cpu)
     { }
     //
     UWORD Execute(UWORD operand)
@@ -246,8 +267,8 @@ public:
     Step3 Third;
     //
   public:
-    Cat3(class CPU *cpu)
-      : First(cpu), Second(cpu), Third(cpu)
+    Cat3(class CPU *cpu,bool haltonrdy = true)
+      : MicroCode(haltonrdy), First(cpu), Second(cpu), Third(cpu)
     { }
     //
     UWORD Execute(UWORD operand)
@@ -262,8 +283,8 @@ public:
     Step First;
     //
   public:
-    Cat1(class CPU *cpu)
-      : First(cpu)
+    Cat1(class CPU *cpu,bool haltonrdy = true)
+      : MicroCode(haltonrdy), First(cpu)
     { }
     //
     UWORD Execute(UWORD operand)
@@ -352,7 +373,7 @@ public:
   // its operand. This is the second step in an (indirect,x) addressing mode.
   // It always adds one additional wait.
   class AddXUnitWait : public AtomicExecutionUnit<class AdrSpace> {
-    T(CPU,Cat1)<WaitUnit> Wait;
+    Cat1<WaitUnit> Wait;
   public:
     AddXUnitWait(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Wait(cpu)
@@ -384,7 +405,7 @@ public:
   // boundary is crossed. This is the second step in an absolute,X addressing
   // step.
   class AddXUnit : public AtomicExecutionUnit<class AdrSpace> {
-    T(CPU,Cat1)<WaitUnit> Wait;
+    Cat1<WaitUnit> Wait;
   public:
     AddXUnit(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Wait(cpu)
@@ -405,7 +426,7 @@ public:
   // its operand. This is the thrid step in an (indirect),y addressing mode
   // and *may* add a wait state.
   class AddYUnitWait : public AtomicExecutionUnit<class AdrSpace> {
-    T(CPU,Cat1)<WaitUnit> Wait;
+    Cat1<WaitUnit> Wait;
   public:
     AddYUnitWait(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Wait(cpu)
@@ -437,7 +458,7 @@ public:
   // its operand. This is the thrid step in an (indirect),y addressing mode
   // and *may* add a wait state.
   class AddYUnit : public AtomicExecutionUnit<class AdrSpace> {
-    T(CPU,Cat1)<WaitUnit> Wait;
+    Cat1<WaitUnit> Wait;
   public:
     AddYUnit(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Wait(cpu)
@@ -494,7 +515,7 @@ public:
   // The 65C02 fixes this behaivour
   class IndirectionUnitExtendFixed : public AtomicExecutionUnit<class AdrSpace> {
     // Used to create a wait-state
-    T(CPU,Cat1)<WaitUnit> Wait;
+    Cat1<WaitUnit> Wait;
   public:
     IndirectionUnitExtendFixed(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Wait(cpu)
@@ -970,7 +991,7 @@ public:
   class ADCUnitFixed : public AtomicExecutionUnit<class AdrSpace> {
     // The wait state we may insert here for the 65C02
     // in decimal mode.
-    T(CPU,Cat1)<WaitUnit> Wait;
+    Cat1<WaitUnit> Wait;
   public:
     ADCUnitFixed(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Wait(cpu)
@@ -999,7 +1020,7 @@ public:
   class SBCUnitFixed : public AtomicExecutionUnit<class AdrSpace> { 
     // The wait state we may insert here for the 65C02
     // in decimal mode.
-    T(CPU,Cat1)<WaitUnit> Wait;
+    Cat1<WaitUnit> Wait;
   public:
     SBCUnitFixed(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Wait(cpu)
@@ -1070,7 +1091,7 @@ public:
   // be inserted directly, but it should be part of the BranchDetectUnit step.
   class BranchUnit : public AtomicExecutionUnit<class AdrSpace> {
     // The wait state we may insert here.
-    T(CPU,Cat1)<WaitUnit> Wait;
+    Cat1<WaitUnit> Wait;
   public:
     BranchUnit(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Wait(cpu)
@@ -1084,7 +1105,7 @@ public:
   // may or may not insert additional steps.
   template<UBYTE mask,UBYTE value>
   class BranchDetectUnit : public AtomicExecutionUnit<class AdrSpace> {
-    T(CPU,Cat1)<BranchUnit> Branch;
+    Cat1<BranchUnit> Branch;
   public:
     BranchDetectUnit(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Branch(cpu)
@@ -1096,8 +1117,8 @@ public:
   // The BranchBitTestUnit, used for the Rockwell BBR/BBS instructions.
   template<UBYTE bitmask,UBYTE bitvalue>
   class BranchBitTestUnit : public AtomicExecutionUnit<class AdrSpace> {
-    T(CPU,Cat1)<BranchUnit> Branch;
-    T(CPU,Cat1)<WaitUnit>   Wait;
+    Cat1<BranchUnit> Branch;
+    Cat1<WaitUnit>   Wait;
   public:
     BranchBitTestUnit(class CPU *cpu)
       : AtomicExecutionUnit<AdrSpace>(cpu), Branch(cpu), Wait(cpu)
@@ -1392,6 +1413,12 @@ public:
   // coldstart.
   void ClearInstructions(void);
   //
+  // Run the microcode decoder, fetch the next instruction.
+  // This is here and not in an atomic step because the monitor
+  // could replace the latter on the fly when inserting
+  // watch-points.
+  UWORD DecodeInstruction(void);
+  //
   // Since the instruction builder is too LONG for some architectures
   // that would require here out-of-range branches, we split it up into
   // several groups.
@@ -1501,39 +1528,40 @@ public:
   // CPU driver to emulate a single cycle of the CPU.
   void Step(void)
   {
+    class MicroCode *current = NextStep; // get the next step to be performed
     //
     // Check whether we have a legal pointer here.
 #if CHECK_LEVEL > 0
     if (size_t(CurCycle - StolenCycles) >= sizeof(StolenCycles)) {
       Throw(OutOfRange,"CPU::Step","execution HPOS out of range");
     }
+    if (current == NULL) {
+      machine->PutWarning("GlobalPC = %04x, IR = %02x\n",GlobalPC,LastIR);
+      Throw(ObjectDoesntExist,"CPU::Step","no current execution step");
+    }
 #endif
     //
     // Check whether there is a CPU slot available (and not stolen by DMA)
     // or blocked by WSync wait
-    if (*CurCycle == 0) {
-      class MicroCode *current = NextStep; // get the next step to be performed
+    if ((*CurCycle & current->StopMask()) == 0) {
       // Get already the step following this step such that the AEU may insert a step
       // between this and the next. This might be required to implement some data
       // dependent delay slots.
       NextStep = *ExecutionSteps;
       ExecutionSteps++;
-#if CHECK_LEVEL > 0
-      if (current == NULL) {
-	machine->PutWarning("GlobalPC = %04x, IR = %02x\n",GlobalPC,LastIR);
-	Throw(ObjectDoesntExist,"CPU::Step","no current execution step");
-      }
-#endif
       AtomicExecutionOperand = current->Execute(AtomicExecutionOperand);
     }
-    CycleCounter++;
     // Bump the horizontal position.
     CurCycle++;
     //CurCycle - StolenCycles
     //
     // Advance the rest of the hardware by a single cycle
-    if (CurCycle <= LastCycle) 
+    if (CurCycle <= LastCycle) {
+      CycleCounter++;
+      // Bump the profile counter
+      ProfileCounter++;
       machine->Step();
+    }
   }
   //
   // Go for an indicated number of cycles and return the
@@ -1575,13 +1603,22 @@ public:
   //
   // Generate a maskable interrupt, given a mask that identifies the source of the
   // IRQ.
-  void GenerateIRQ(ULONG devicemask);
+  void GenerateIRQ(ULONG devicemask)
+  {
+    IRQMask  |= devicemask;
+  }
   //
   // Release the IRQ from a given device, identified by this mask.
-  void ReleaseIRQ(ULONG devicemask);
+  void ReleaseIRQ(ULONG devicemask)
+  {
+    IRQMask &= ~devicemask;
+  }
   //
   // Generate a non-maskable interrupt
-  void GenerateNMI(void);
+  void GenerateNMI(void)
+  {
+    NMI = true; // that's all!
+  }
   //
   // Generate a watch point interrupt.
   void GenerateWatchPoint(UBYTE idx)
@@ -1624,6 +1661,26 @@ public:
   {
     return Emulate65C02;
   }
+  //
+  // Start profiling CPU execution.
+  void StartProfiling(void);
+  //
+  // Stop profiling CPU execution.
+  void StopProfiling(void);
+  //
+  // Return the profile conters, one per possible PC location.
+  // Returns NULL in case profiling is disabled.
+  const ULONG *ProfilingCountersOf(void) const
+  {
+    return ProfilingCounters;
+  }
+  //
+  // Return the cumulative profiling counters that accumulate
+  // over subroutine calls.
+  const ULONG *CumulativeProfilingCountersOf(void) const
+  {
+    return CumulativeCounters;
+  }
 };
 ///
 
