@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: hdevice.cpp,v 1.44 2009-05-16 20:00:28 thor Exp $
+ ** $Id: hdevice.cpp,v 1.55 2021/07/03 16:16:44 thor Exp $
  **
  ** In this module: H: emulated device for emulated disk access.
  **********************************************************************************/
@@ -52,7 +52,7 @@ HDevice::~HDevice(void)
 /// HDevice::BinaryLoadCallbackPatch::BinaryLoadCallbackPatch
 // The callback patch of the binary load interface.
 HDevice::BinaryLoadCallbackPatch::BinaryLoadCallbackPatch(class Machine *mach,class PatchProvider *p)
-  : Patch(mach,p,1), State(Idle), CPU(mach->CPU())
+  : Patch(mach,p,1), State(Idle), CPU(mach->CPU()), machine(mach)
 {
 }
 ///
@@ -134,8 +134,10 @@ void HDevice::BinaryLoadCallbackPatch::RunPatch(class AdrSpace *,class CPU *,UBY
       // Check whether we already have a run address.
       // If not, then keep this initial address as
       // run address for OS/A+ compatibility.
+      /* BAD idea, not Atari compatible
       BaseSpace->WriteByte(0x2e0,Start & 0xff); // lo
       BaseSpace->WriteByte(0x2e1,Start >> 8  ); // hi
+      */
       // Fine. Keep it.
       State = ReadEnd;
       ReadBlock(0x240,0x02);
@@ -178,19 +180,18 @@ void HDevice::BinaryLoadCallbackPatch::RunPatch(class AdrSpace *,class CPU *,UBY
       break;
     } else if (ErrorCode == 0x03) {
       UWORD init;
+      UWORD run;
       // We are at the EOF. That's fine and no
       // error.
       ErrorCode = 0x01;      
-      // At the end, call us again to close up.
-      State     = Run;
-      PushReturn(0xe47a);
       // Keep the channel
       // open, and jump thru the run address
-      // should we have been asked for.
-      if (RunMask & 0x40) {
+      // should we have been asked for. 
+      run = BaseSpace->ReadWord(0x2e0);
+      if (run && (RunMask & 0x40)) {
 	// Read it from memory for maximum
 	// compatibility.
-	PushReturn(BaseSpace->ReadWord(0x2e0));
+	PushReturn(run);
       }
       //
       // Call the init vector first if we have one.
@@ -199,6 +200,9 @@ void HDevice::BinaryLoadCallbackPatch::RunPatch(class AdrSpace *,class CPU *,UBY
 	// Call it.
 	PushReturn(init);
       }
+      //
+      // But even before that, close the channel.
+      CloseChannel();
       break;
     }
     // Otherwise, close the channel and return.
@@ -261,6 +265,11 @@ void HDevice::BinaryLoadCallbackPatch::RunPatch(class AdrSpace *,class CPU *,UBY
     //
     // Run into the following.
   case Idle:
+    // Generate a warning.
+    machine->PutWarning("A program is currently trying to boot "
+			"from the tape, however the tape is currently "
+			"unavailable due to the HDevice patch. Disable the "
+			"HDevice patch in the OsRom menu to allow booting.");
     // If we are called here, then a program called the cassette init vector
     // without really knowing what is going on. Return with an error code.
     CPU->Y()  = 0x90;
@@ -353,11 +362,31 @@ void HDevice::BinaryLoadCallbackPatch::LaunchBinaryLoad(class AdrSpace *adr,UBYT
 // Check whether a file name matches a pattern
 bool HDevice::HandlerChannel::MatchPattern(const char *filename,char *pattern)
 {
+  const char *ch = filename;
+  bool havedot   = false;
+  //
   // If the file name contains a '.', we never match. Hidden files
   // are never shown here.
   if (*filename == '.')
     return false;
   //
+  // If the filename contains something that is not printable, do not
+  // match either.
+  while(*ch) {
+    if (*ch <= 0x20 || 
+	(*ch >= 0x21 && *ch <= 0x40 && *ch != '.' && !(*ch >= '0' && *ch <= '9')) ||
+	(*ch >= 0x5b && *ch <= 0x60) || (*ch >= 0x7b)) {
+      if (havedot == false || *ch != ' ') {
+	return false;
+      }
+    }
+    if (*ch == '.') {
+      if (havedot)
+	return false;
+      havedot = true;
+    }
+    ch++;
+  }
   // If the pattern contains a '-', we always match. (Thor-Dos)
   // We do not check trailing characters here.
   if (*pattern == '-')
@@ -569,14 +598,15 @@ UBYTE HDevice::HandlerChannel::ToDirEntry(void)
 {
   size_t i,len;
   struct stat info;
-  char fullname[256];
+  char fullname[257];
   const char *name = de_name(fib);
   const char *dot;
 
   if (strlen(basedir) + 3 + strlen(name) > 256) {
     return 0xa5;
   }
-  snprintf(fullname,255,"%s/%s",basedir,name);
+  if (snprintf(fullname,sizeof(fullname),"%s/%s",basedir,name) > int(sizeof(fullname)))
+    return 0xa5; // invalid file name
   if (stat(fullname,&info) == -1) {
     return AtariError(errno);
   }
@@ -632,6 +662,11 @@ void HDevice::FilterAux1(char *name,UBYTE &aux1)
 {
   char *modifier;
   //
+  // Check whether there is a comma in the name, remove it.
+  modifier = strchr(name,',');
+  if (modifier) // cut off.
+    *modifier = 0;
+  //
   // Scan the file name/pattern for the "/" modifier indicator.
   modifier = strchr(name,'/');
   if (modifier) {
@@ -941,6 +976,7 @@ UBYTE HDevice::Special(UBYTE channel,UBYTE unit,class AdrSpace *adr,UBYTE cmd,
   UBYTE result = 0x01;
   struct HandlerChannel *ch;
   char path[256];
+  char *comma;
   //
   // Whether we need a channel here or not depends on the specific CIO command.
   switch (cmd) {
@@ -967,10 +1003,13 @@ UBYTE HDevice::Special(UBYTE channel,UBYTE unit,class AdrSpace *adr,UBYTE cmd,
     ch          = new struct HandlerChannel(0x00,BaseDir[unit-1]);
     Buffer[8]   = ch;
     channel     = 8;
-    ExtractFileName(adr,mem,path,sizeof(path));    
+    ExtractFileName(adr,mem,path,sizeof(path));
+    comma       = strchr(path,',');
     FilterAux1(path,aux[0]);
     switch(cmd) {
     case 0x20:
+      if (comma)
+	*comma = ',';
       result  = Rename(ch,path);
       break;
     case 0x21:
@@ -1109,7 +1148,8 @@ UBYTE HDevice::Rename(struct HandlerChannel *ch,char *pattern)
     return 0xa5;
   //
   // Build the destination name
-  snprintf(destpath,255,"%s/%s",ch->basedir,dst);
+  if (snprintf(destpath,sizeof(destpath),"%s/%s",ch->basedir,dst) > int(sizeof(destpath)))
+    return 0xa5; // Invalid file name
 
   result = ch->MatchFirst(pattern);
   while(result == 0x01) {
@@ -1124,7 +1164,8 @@ UBYTE HDevice::Rename(struct HandlerChannel *ch,char *pattern)
       }
     }
     // Worked so far. Now run the rename.
-    snprintf(sourcepath,255,"%s/%s",ch->basedir,de_name(ch->fib));
+    if (snprintf(sourcepath,sizeof(sourcepath),"%s/%s",ch->basedir,de_name(ch->fib)) > int(sizeof(sourcepath)))
+      return 0xa5; // invalid file name
     if (rename(sourcepath,destpath) < 0) {
       return ch->AtariError(errno);
     }
@@ -1148,7 +1189,8 @@ UBYTE HDevice::Delete(struct HandlerChannel *ch,char *pattern)
 
   result = ch->MatchFirst(pattern);
   while (result == 0x01) {
-    snprintf(target,255,"%s/%s",ch->basedir,de_name(ch->fib));
+    if (snprintf(target,sizeof(target),"%s/%s",ch->basedir,de_name(ch->fib)) > int(sizeof(target)))
+      return 0xa5; // Invalid file name
     // Now check whether this file is "protected". If so,
     // deliver an error and do not delete it.
     if (stat(target,&info) == -1) {
@@ -1239,7 +1281,8 @@ UBYTE HDevice::Protect(struct HandlerChannel *ch,char *pattern)
 
   result = ch->MatchFirst(pattern);
   while (result == 0x01) {
-    snprintf(target,255,"%s/%s",ch->basedir,de_name(ch->fib));
+    if (snprintf(target,sizeof(target),"%s/%s",ch->basedir,de_name(ch->fib)) > int(sizeof(target)))
+      return 0xa5; // Invalid file name
     // Get the info for this file.
     if (stat(target,&info) < 0) {
       return ch->AtariError(errno);
@@ -1269,7 +1312,8 @@ UBYTE HDevice::Unprotect(struct HandlerChannel *ch,char *pattern)
 
   result = ch->MatchFirst(pattern);
   while (result == 0x01) {
-    snprintf(target,255,"%s/%s",ch->basedir,de_name(ch->fib));
+    if (snprintf(target,sizeof(target),"%s/%s",ch->basedir,de_name(ch->fib)) > int(sizeof(target)))
+      return 0xa5; // Invalid file name
     // Get the info for this file.
     if (stat(target,&info) < 0) {
       return ch->AtariError(errno);
