@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: x11_truecolorbuffer.cpp,v 1.15 2008-05-22 13:03:54 thor Exp $
+ ** $Id: x11_truecolorbuffer.cpp,v 1.18 2021/08/16 10:31:01 thor Exp $
  **
  ** In this module: Conversions from ANTIC/GTIA output to X11 draw commands
  **********************************************************************************/
@@ -101,7 +101,8 @@ void X11_TrueColorBuffer::ScanBlock::AddPixel(int x,int y,int w, int h)
       // Yup, connects fine to it. Just enlarge me horizontally
       last->width += w;
       return;
-    }
+    } 
+    //
     // Now test whether we have pixels one row above
     while(test >= 0) {
       int lastend = last->y + last->height;
@@ -110,7 +111,6 @@ void X11_TrueColorBuffer::ScanBlock::AddPixel(int x,int y,int w, int h)
 	last->height += h;
 	return;
       }
-      if (lastend < y) break; // cannot match anymore, these are sorted
       last--;
       test--;
     }
@@ -160,7 +160,7 @@ void X11_TrueColorBuffer::RebuildScreen(bool differential)
   if (indexdirty) {
     PackedRGB *row;
     UBYTE     *idxrow;
-    const struct GTIA::ColorEntry *colormap = machine->GTIA()->ActiveColorMap();
+    const struct ColorEntry *colormap = machine->GTIA()->ActiveColorMap();
     LONG       height = Height;
     //
     // Compute the start addresses of the row/packed row.
@@ -576,13 +576,13 @@ void X11_TrueColorBuffer::DisplayStatus(class Monitor *mon)
 {
   mon->PrintStatus("X11_TrueColorBuffer status:\n"
 		   "\tIndirect rendering    : %s\n"
-		   "\tPixel width           : " LD "\n"
-		   "\tPixel height          : " LD "\n"
+		   "\tPixel width           : " ATARIPP_LD "\n"
+		   "\tPixel height          : " ATARIPP_LD "\n"
 		   "\tTrue Color Renderer   : on\n"
-		   "\tLeftEdge              : " LD "\n"
-		   "\tTopEdge               : " LD "\n"
-		   "\tWidth                 : " LD "\n"
-		   "\tHeight                : " LD "\n",
+		   "\tLeftEdge              : " ATARIPP_LD "\n"
+		   "\tTopEdge               : " ATARIPP_LD "\n"
+		   "\tWidth                 : " ATARIPP_LD "\n"
+		   "\tHeight                : " ATARIPP_LD "\n",
 		   (PixMapIndirect)?("on"):("off"),
 		   PixelWidth,PixelHeight,
 		   LeftEdge,TopEdge,Width,Height);
@@ -597,7 +597,7 @@ void X11_TrueColorBuffer::PushLine(UBYTE *in,int size)
   PackedRGB *outbuf = row;
   PackedRGB *out    = outbuf;
   int i             = size;  
-  const struct GTIA::ColorEntry *colormap = machine->GTIA()->ActiveColorMap();
+  const struct ColorEntry *colormap = machine->GTIA()->ActiveColorMap();
 
   if (outbuf) {
     // Convert this line to RGB, then push into the
